---------------------------------------------------------------------------------------------------
--   Copyright (c) 2001 R.A.Matthews
--
--   See end of file for licence notice.
--
--   Revision History:
--      15 November 2001 R.A.Matthews
--         Derived from bm.List_Handling (part of the Gapse project).
--         Removed 'bm.Error_Handling' exception handling facilities.
--
---------------------------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;

package body Doubly_Linked_List is

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Prepend (List  : in out List_Type;
                      Data  : in     Data_Type)
   is
   
   begin
      if Is_Empty (List) then
      
         List.Header.First := new Data_Envelope_Type'(Data     => new Data_Type'(Data),
                                                      Next     => null,
                                                      Previous => null);
         List.Header.Last := List.Header.First;
      else
         List.Header.First.Previous := new Data_Envelope_Type'(Data     => new Data_Type'(Data),
                                                               Next     => List.Header.First,
                                                               Previous => null);
         List.Header.First := List.Header.First.Previous;
      end if;

      --  Any enumerators that were denoting the start of the list must be updated to indicate that
      --  they now follow the new data element...
      
      declare
         Current_Enumerator : Enumerator_Envelope_Access_Type;
      begin
         Current_Enumerator := List.Header.Enumerators;
         loop
            exit when Current_Enumerator = null;     -- ****************
            
            if Current_Enumerator.Previous_Data = null then
            
               Current_Enumerator.Previous_Data := List.Header.First;
            end if;
            
            Current_Enumerator := Current_Enumerator.Next_Enumerator;
         end loop;
      end;

   end Prepend;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Append (List  : in out List_Type;
                     Data  : in     Data_Type)
   is
   
   begin
      if Is_Empty (List) then
      
         List.Header.First := new Data_Envelope_Type'(Data     => new Data_Type'(Data),
                                                      Next     => null,
                                                      Previous => null);
         List.Header.Last := List.Header.First;
      else
         List.Header.Last.Next := new Data_Envelope_Type'(Data     => new Data_Type'(Data),
                                                          Next     => null,
                                                          Previous => List.Header.Last);
         List.Header.Last := List.Header.Last.Next;
      end if;

      --  Any enumerators that were denoting the end of the list must be updated to indicate that
      --  they now precede the new data element...
      
      declare
         Current_Enumerator : Enumerator_Envelope_Access_Type;
      begin
         Current_Enumerator := List.Header.Enumerators;
         loop
            exit when Current_Enumerator = null;     -- ****************
            
            if Current_Enumerator.Next_Data = null then
            
               Current_Enumerator.Next_Data := List.Header.Last;
            end if;
            
            Current_Enumerator := Current_Enumerator.Next_Enumerator;
         end loop;
      end;

   end Append;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Delete_All (List  : in out List_Type)
is
   
      procedure Free is new Ada.Unchecked_Deallocation (Object => Data_Type,
                                                        Name   => Data_Access_Type);

      procedure Free is new Ada.Unchecked_Deallocation (Object => Data_Envelope_Type,
                                                        Name   => Data_Envelope_Access_Type);
      
      Current_Enumerator    : Enumerator_Envelope_Access_Type;
      Current_Data_Envelope : Data_Envelope_Access_Type;
      
   begin
      --  Update the enumerators (if any) to indicate that the data has been deleted...
      
      Current_Enumerator := List.Header.Enumerators;
      
      loop
         exit when Current_Enumerator = null;   -- **********************
         
         Current_Enumerator.Current             := null;
         Current_Enumerator.Next_Data           := null;
         Current_Enumerator.Previous_Data       := null;
         
         Current_Enumerator := Current_Enumerator.Next_Enumerator;
      end loop;

      --  Deallocate each data element (if any) and its envelope...
      
      Current_Data_Envelope := List.Header.First;
      
      loop
         exit when Current_Data_Envelope = null;   -- ******************
         
         Free (Current_Data_Envelope.Data);
         
         if Current_Data_Envelope.Next = null then
         
            Free (Current_Data_Envelope);   -- Sets Current_Data_Envelope to null.
         else
            Current_Data_Envelope := Current_Data_Envelope.Next;
            Free (Current_Data_Envelope.Previous);
         end if;
      end loop;

      List.Header.First := null;
      List.Header.Last  := null;

   end Delete_All;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function Is_Empty (List : in List_Type) return Boolean
is
   
   begin

      return (List.Header.First = null);
      
   end Is_Empty;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function  "=" (Left  : in List_Type;
                  Right : in List_Type) return Boolean
   is
   
      Left_Enum  : Enumerator_Type;
      Right_Enum : Enumerator_Type;
      
   begin
      if Is_Empty (Left) and Is_Empty (Right) then
      
         return True;
         --  ****************************************************************************************
      elsif Is_Empty (Left) or Is_Empty (Right) then
      
         return False;
         --  ****************************************************************************************
      else
         Left_Enum  := New_Enumerator (Left);
         Right_Enum := New_Enumerator (Right);
         
         Goto_First (Left_Enum);
         Goto_First (Right_Enum);
         loop
            if Current (Left_Enum) /= Current (Right_Enum) then
            
               return False;
               --  **********************************************************************************
            else
               if Is_Last (Left_Enum) and Is_Last (Right_Enum) then
               
                  return True;
                  --  *******************************************************************************
               elsif Is_Last (Left_Enum) or Is_Last (Right_Enum) then
               
                  return False;
                  --  *******************************************************************************
               else
                  Goto_Next (Left_Enum);
                  Goto_Next (Right_Enum);
               end if;
            end if;
         end loop;
      end if;
      
   end "=";

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function New_Enumerator (List : in List_Type'Class) return Enumerator_Type is
   
      Enum : Enumerator_Envelope_Access_Type;
      
   begin
      --  Create the new enumerator and insert it in the list's enumerator sub-list...
      
      Enum := new Enumerator_Envelope_Type'(List_Header         => List.Header,
                                            Current             => List.Header.First,
                                            Next_Enumerator     => List.Header.Enumerators,
                                            Previous_Enumerator => null,
                                            Next_Data           => null,
                                            Previous_Data       => null);

      if Enum.Current /= null then
      
         Enum.Next_Data     := Enum.Current.Next;
         Enum.Previous_Data := Enum.Current.Previous;
      end if;

      if Enum.Next_Enumerator /= null then
      
         Enum.Next_Enumerator.Previous_Enumerator := Enum;
      end if;
      
      List.Header.Enumerators := Enum;

      return (Ada.Finalization.Controlled with Header => Enum);
      
   end New_Enumerator;
   
   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Insert (Enumerator : in out Enumerator_Type;
                     Data       : in     Data_Type;
                     Position   : in     Position_Type)
   is
   
      New_Data_Element : Data_Envelope_Access_Type;
      
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************
      end if;

      ---------------------------------------------------------------------------------------------
      --  This routine operates in three phases:
      --  1) Insert the data into the list associated with the enumerator.
      --  2) Update all enumerators associated with the list, to reflect the existence of the new
      --     data element.
      --  3) Set the new data element as current for the enumerator passed as the call parameter.
      ---------------------------------------------------------------------------------------------
      --
      --  Phase 1: insert the data into the list...
      
      if Current_Exists (Enumerator) then
      
         case Position is
            when Before_Current =>
               New_Data_Element := new Data_Envelope_Type'(Data     => new Data_Type'(Data),
                                                           Next     => Enumerator.Header.Current,
                                                           Previous => Enumerator.Header.Previous_Data);
            when After_Current =>
               New_Data_Element := new Data_Envelope_Type'(Data     => new Data_Type'(Data),
                                                           Next     => Enumerator.Header.Next_Data,
                                                           Previous => Enumerator.Header.Current);
         end case;
      else
         New_Data_Element := new Data_Envelope_Type'(Data     => new Data_Type'(Data),
                                                     Next     => Enumerator.Header.Next_Data,
                                                     Previous => Enumerator.Header.Previous_Data);
      end if;
      
      if New_Data_Element.Next = null then
      
         --  The data must be placed at the end of the list.
         Enumerator.Header.List_Header.Last := New_Data_Element;
      else
         New_Data_Element.Next.Previous := New_Data_Element;
      end if;
      
      if New_Data_Element.Previous = null then
      
         --  The data must be placed at the start of the list.
         Enumerator.Header.List_Header.First := New_Data_Element;
      else
         New_Data_Element.Previous.Next := New_Data_Element;
      end if;

      
      --  Phase 2: update the enumerators.
      --
      --  Each enumerator whose Next_Data and Previous_Data components now straddle the new data element
      --  must be amended.
      --  Note that particular care is required for those enumerators whose current has been deleted. Here
      --  we must determine whether the new data element precedes or follows the enumerator...
      
      declare
         Current_Enumerator_Header : Enumerator_Envelope_Access_Type;
      begin
         Current_Enumerator_Header := Enumerator.Header.List_Header.Enumerators;
         loop
            exit when Current_Enumerator_Header = null;                      -- ****************
            
            if Current_Enumerator_Header.Current = null then
            
               --  The current enumerator's current data element has been deleted.
               if New_Data_Element.Previous = Current_Enumerator_Header.Previous_Data then
               
                  case Position is
                     when Before_Current =>
                        if Enumerator.Header.Current = null then
                        
                           --  The enumerator used for the insert is at the same position as the current
                           --  enumerator.
                           Current_Enumerator_Header.Previous_Data := New_Data_Element;
                        else
                           --  The enumerator used for the insert has a current that immediately follows
                           --  the current enumerator.
                           Current_Enumerator_Header.Next_Data     := New_Data_Element;
                        end if;
                     when After_Current =>
                        if Enumerator.Header.Current = null then
                        
                           --  The enumerator used for the insert is at the same position as the current
                           --  enumerator.
                           Current_Enumerator_Header.Next_Data     := New_Data_Element;
                        else
                           --  The enumerator used for the insert has a current that immediately precedes
                           --  the current enumerator.
                           Current_Enumerator_Header.Previous_Data := New_Data_Element;
                        end if;
                  end case;
               else
                  null; -- The new data element does not affect the current enumerator.
               end if;
            else
               if New_Data_Element.Previous = Current_Enumerator_Header.Previous_Data then
               
                  --  New element is just before current.
                  Current_Enumerator_Header.Previous_Data := New_Data_Element;
               
               elsif New_Data_Element.Next = Current_Enumerator_Header.Next_Data then
               
                  --  New element is just after current.
                  Current_Enumerator_Header.Next_Data := New_Data_Element;
               else
                  null; -- The new data element does not affect the current enumerator.
               end if;
            end if;
            
            Current_Enumerator_Header := Current_Enumerator_Header.Next_Enumerator;
         end loop;
      end;
      
      
      --  Phase 3: we must position the enumerator, passed as our call parameter, at the new
      --  data element...
      
      Enumerator.Header.Current       := New_Data_Element;
      Enumerator.Header.Previous_Data := New_Data_Element.Previous;
      Enumerator.Header.Next_Data     := New_Data_Element.Next;
      
   end Insert;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Delete (Enumerator : in Enumerator_Type)
is

      procedure Free is new Ada.Unchecked_Deallocation (Object => Data_Type,
                                                        Name   => Data_Access_Type);

      procedure Free is new Ada.Unchecked_Deallocation (Object => Data_Envelope_Type,
                                                        Name   => Data_Envelope_Access_Type);

   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************
      
      elsif not Current_Exists (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************

      else
         --  First we remove the current data element from the list...
         
         if Is_First (Enumerator) and Is_Last (Enumerator) then
         
            --  We are deleting with the only element in the list.
            Enumerator.Header.List_Header.First := null;
            Enumerator.Header.List_Header.Last  := null;

         elsif Is_First (Enumerator) then
         
            --  We have a list with two or more elements, we are deleting the first.
            Enumerator.Header.List_Header.First          := Enumerator.Header.Current.Next;
            Enumerator.Header.List_Header.First.Previous := null;
         
         elsif Is_Last (Enumerator) then
         
            --  We have a list with two or more elements, we are deleting the last.
            Enumerator.Header.List_Header.Last      := Enumerator.Header.Current.Previous;
            Enumerator.Header.List_Header.Last.Next := null;
         
         else
            --  We are deleting an element with both next and previous elements.
            Enumerator.Header.Current.Previous.Next := Enumerator.Header.Current.Next;
            Enumerator.Header.Current.Next.Previous := Enumerator.Header.Current.Previous;
         end if;

         --  Now we must remove the data element from the all enumerators associated with the list.
         --  Note that this includes the enumerator we have as a call parameter, hence our use of
         --  a local variable to denote the data element we are deleting.
         declare
            Current_Data_Envelope : Data_Envelope_Access_Type;
            Current_Enumerator    : Enumerator_Envelope_Access_Type;
         begin
            Current_Data_Envelope := Enumerator.Header.Current;

            Current_Enumerator := Enumerator.Header.List_Header.Enumerators;
            loop
               exit when Current_Enumerator = null;   --  ****************
               
               if Current_Enumerator.Current = Current_Data_Envelope then
               
                  Current_Enumerator.Current := null;
                  
               elsif Current_Enumerator.Next_Data = Current_Data_Envelope then
               
                  Current_Enumerator.Next_Data := Current_Data_Envelope.Next;
                  
               elsif Current_Enumerator.Previous_Data = Current_Data_Envelope then
               
                  Current_Enumerator.Previous_Data := Current_Data_Envelope.Previous;
               end if;
               
               Current_Enumerator := Current_Enumerator.Next_Enumerator;
            end loop;

            --  Finally we deallocate the data element...
            
            Free (Current_Data_Envelope.Data);
            Free (Current_Data_Envelope);
         end;
      end if;
      
   end Delete;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function Current (Enumerator : in Enumerator_Type) return Data_Type is
  
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************
      
      elsif not Current_Exists (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************
      
      else
         return Enumerator.Header.Current.Data.all;
      end if;
      
   end Current;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Update (Enumerator : in Enumerator_Type;
                     Data       : in Data_Type)
   is
   
      procedure Free is new Ada.Unchecked_Deallocation (Object => Data_Type,
                                                        Name   => Data_Access_Type);
      
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************
      
      elsif not Current_Exists (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************

      else
         begin
            Enumerator.Header.Current.Data.all := Data;
            --  Note that new data must be compatible with the old, otherwise Ada run-time will raise Contraint_Error.
         exception
            when Constraint_Error =>
               --  The new data did not match the old, so we must create a new data element to replace the old...
               
               --  First delete the old:
               
               Free (Enumerator.Header.Current.Data);
               
               --  Now create the new:
               
               Enumerator.Header.Current.Data := new Data_Type'(Data);
         end;
      end if;

   end Update;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Goto_First (Enumerator : in out Enumerator_Type)
is
   
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      elsif Is_Empty (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************

      else
         Enumerator.Header.Current       := Enumerator.Header.List_Header.First;
         Enumerator.Header.Next_Data     := Enumerator.Header.Current.Next;
         Enumerator.Header.Previous_Data := Enumerator.Header.Current.Previous;
      end if;
      
   end Goto_First;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Goto_Last (Enumerator : in out Enumerator_Type) is
   
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      elsif Is_Empty (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************

      else
         Enumerator.Header.Current       := Enumerator.Header.List_Header.Last;
         Enumerator.Header.Next_Data     := Enumerator.Header.Current.Next;
         Enumerator.Header.Previous_Data := Enumerator.Header.Current.Previous;
      end if;
      
   end Goto_Last;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Goto_Next (Enumerator : in out Enumerator_Type)
is

   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      elsif Is_Empty (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************
      
      elsif Is_Last (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************

      else
         --  At this point: the enumerator's list exists, the list is not empty, and at least
         --  one data element follows the current position. Note that the current element may
         --  or may not exist.
         Enumerator.Header.Current       := Enumerator.Header.Next_Data;
         Enumerator.Header.Next_Data     := Enumerator.Header.Current.Next;
         Enumerator.Header.Previous_Data := Enumerator.Header.Current.Previous;
      end if;
      
   end Goto_Next;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Goto_Previous (Enumerator : in out Enumerator_Type) is
   
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      elsif Is_Empty (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************
      
      elsif Is_First (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************

      else
         --  At this point: the enumerator's list exists, the list is not empty, and at least
         --  one data element precedes the current position. Note that the current element may
         --  or may not exist.
         Enumerator.Header.Current       := Enumerator.Header.Previous_Data;
         Enumerator.Header.Next_Data     := Enumerator.Header.Current.Next;
         Enumerator.Header.Previous_Data := Enumerator.Header.Current.Previous;
      end if;
      
   end Goto_Previous;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function Is_First (Enumerator : in Enumerator_Type) return Boolean
is

   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      elsif Is_Empty (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************
      
      else
         --  At this point the enumerator's list exists and it is not empty.
         --  Note that the current element may or may not exist.
         return (Enumerator.Header.Previous_Data = null);
      end if;
      
   end Is_First;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function Is_Last (Enumerator : in Enumerator_Type) return Boolean
is
   
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      elsif Is_Empty (Enumerator) then
      
         raise No_Data;
         --  ****************************************************************************************
      
      else
         --  At this point the enumerator's list exists and it is not empty.
         --  Note that the current element may or may not exist.
         
         return (Enumerator.Header.Next_Data = null);
      end if;
      
   end Is_Last;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function Current_Exists (Enumerator : in Enumerator_Type) return Boolean
is

   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      else
         return (Enumerator.Header.Current /= null);
      end if;
      
   end Current_Exists;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function List_Exists (Enumerator : in Enumerator_Type) return Boolean
is

   begin

      return (Enumerator.Header.List_Header /= null);

   end List_Exists;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function Is_Empty (Enumerator : in Enumerator_Type) return Boolean is
 
   begin
      if not List_Exists (Enumerator) then
      
         raise No_List;
         --  ****************************************************************************************

      else
         return (Enumerator.Header.List_Header.First = null);
      end if;
      
   end Is_Empty;
 
   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   function "=" (Left  : in Enumerator_Type;
                 Right : in Enumerator_Type) return Boolean
   is

   begin
      return ((Left.Header.List_Header   = Right.Header.List_Header) and then
              (Left.Header.Current       = Right.Header.Current)     and then
              (Left.Header.Next_Data     = Right.Header.Next_Data)   and then
              (Left.Header.Previous_Data = Right.Header.Previous_Data));
   
   end "=";

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Initialize (List : in out List_Type) is

   begin
      List.Header := new List_Header_Type'(First       => null,
                                           Last        => null,
                                           Enumerators => null);

   end Initialize;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Adjust (List : in out List_Type) is

      Temp_List  : List_Type;
      Enumerator : Enumerator_Type;
      
   begin
      --  On entry List.Header points to a List_Header_Type record used by another list. We need
      --  give it its own header and its own data elements.
      --  Note that enumerators are not transferred nor copied in this operation.
      
      if not Is_Empty (List) then
      
         Enumerator := New_Enumerator (List);
         
         Goto_First (Enumerator);
         loop
            Append (List => Temp_List, Data => Current (Enumerator));
            
            exit when Is_Last (Enumerator);   -- ************
            
            Goto_Next (Enumerator);
         end loop;
      end if;
      --  At this point Temp_List contains a copy of all the data elements in List and has those data elements
      --  in the same order as List.
      
      --  We now set List to point to the new list...
      
      List.Header := Temp_List.Header;
      
      --  Finally we must ensure that Temp_List does not point to the new list, otherwise
      --  when it goes out of scope, it will be finalized and the new list will be deleted!
      
      Temp_List.Header   := null;
      
   end Adjust;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Finalize (List : in out List_Type) is

      procedure Free is new Ada.Unchecked_Deallocation (Object => List_Header_Type,
                                                        Name   => List_Header_Access_Type);

      Current_Enumerator : Enumerator_Envelope_Access_Type;
      
   begin
      if List.Header = null then
      
         null;  -- Nothing to do!
      else
         --  Deallocate the data in the list...
         
         Delete_All (List);

         --  Decouple the enumerators (if any) from the list and from one another...
         
         Current_Enumerator := List.Header.Enumerators;
         List.Header.Enumerators := null;
         
         loop
            exit when Current_Enumerator = null;   --  ***********************
            
            Current_Enumerator.List_Header         := null;
            Current_Enumerator.Current             := null;
            Current_Enumerator.Next_Data           := null;
            Current_Enumerator.Previous_Data       := null;
            Current_Enumerator.Previous_Enumerator := null;
            
            Current_Enumerator := Current_Enumerator.Next_Enumerator;
           
            if Current_Enumerator /= null then
            
               Current_Enumerator.Previous_Enumerator.Next_Enumerator := null;
            end if;
         end loop;

         --  Now we can deallocate the list's header...
         
         Free (List.Header);
      end if;

   end Finalize;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure List_Write (Stream : access Ada.Streams.Root_Stream_Type'Class; Item : in  List_Type) is
   
      Enumerator : Enumerator_Type;
   
   begin
      Enumerator := New_Enumerator (Item);
      
      if Is_Empty (Enumerator) then
      
         Boolean'Write (Stream, False);   --  Indicates that no data follows.
      else
         loop
            Boolean'Write (Stream, True);   --  Indicates that data follows.
            
            Data_Type'Output (Stream, Current (Enumerator));
            
            exit when Is_Last (Enumerator);   --  ***********************
            
            Goto_Next (Enumerator);
         end loop;
         
         Boolean'Write (Stream, False);   --  Indicates that no data follows.
         
      end if;

   end List_Write;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure List_Read  (Stream : access Ada.Streams.Root_Stream_Type'Class; Item : out List_Type) is

      Data_Follows : Boolean;
   
   begin
      Delete_All (Item);    --  Delete anything already in the list.
      
      loop
         Boolean'Read (Stream, Data_Follows);
         
         exit when not Data_Follows;    --  ************************
         
         declare
            Data : Data_Type := Data_Type'Input (Stream);
         begin
            Append (List => Item, Data => Data);
         end;
      end loop;

   end List_Read;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Initialize (Enumerator : in out Enumerator_Type) is

   begin
      Enumerator.Header := new Enumerator_Envelope_Type'(List_Header         => null,
                                                         Current             => null,
                                                         Next_Enumerator     => null,
                                                         Previous_Enumerator => null,
                                                         Next_Data           => null,
                                                         Previous_Data       => null);

   end Initialize;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Adjust (Enumerator : in out Enumerator_Type) is

      Temp_Enumerator : Enumerator_Type;
      
   begin
      --  On entry Enumerator points to a Enumerator_Envelope_Type record that is used by another
      --  enumerator. We need to give it its own envelope which we will do via a temporary local
      --  enumerator...
      
      Temp_Enumerator.Header.List_Header         := Enumerator.Header.List_Header;
      Temp_Enumerator.Header.Current             := Enumerator.Header.Current;
      Temp_Enumerator.Header.Next_Data           := Enumerator.Header.Next_Data;
      Temp_Enumerator.Header.Previous_Data       := Enumerator.Header.Previous_Data;
      
      if List_Exists (Temp_Enumerator) then
      
         --  We need to prepend the new enumerator envelope to the associated list's sub-list...
         
         Temp_Enumerator.Header.Next_Enumerator     := Enumerator.Header.List_Header.Enumerators;
         Temp_Enumerator.Header.Previous_Enumerator := null;
      
         Enumerator.Header.List_Header.Enumerators.Previous_Enumerator := Temp_Enumerator.Header;
         Enumerator.Header.List_Header.Enumerators                     := Temp_Enumerator.Header;
      end if;
      
      --  We now set Enumerator to point to the new envelope...
      
      Enumerator.Header := Temp_Enumerator.Header;
      
      --  Finally we must ensure that Temp_Enumerator does not point to the new envelope, otherwise
      --  when it goes out of scope, it will be finalized and the new envelope will be deleted!
      
      Temp_Enumerator.Header   := null;
      
   end Adjust;

   ------------------------------------------------------------------------------------------------
   ------------------------------------------------------------------------------------------------

   procedure Finalize (Enumerator : in out Enumerator_Type) is

      procedure Free is new Ada.Unchecked_Deallocation (Object => Enumerator_Envelope_Type,
                                                        Name   => Enumerator_Envelope_Access_Type);
      
   begin
      if Enumerator.Header = null then
      
         null;   --  Nothing to do!
      else
         if Enumerator.Header.List_Header /= null then
         
            --  We must remove the enumerator from the associated list's enumerator sub-list...
         
            if (Enumerator.Header.Previous_Enumerator = null) and
               (Enumerator.Header.Next_Enumerator     = null)
            then
               --  The enumerator is the only one in the sub-list.
               Enumerator.Header.List_Header.Enumerators := null;
            
            elsif Enumerator.Header.Previous_Enumerator = null then

               --  The enumerator is the first in the sub-list.
               Enumerator.Header.List_Header.Enumerators := Enumerator.Header.Next_Enumerator;
               Enumerator.Header.Next_Enumerator.Previous_Enumerator := null;

            elsif Enumerator.Header.Next_Enumerator = null then
            
               --  The enumerator is the last in the sub-list.
               Enumerator.Header.Previous_Enumerator.Next_Enumerator := null;
            else
               --  The enumerator has a successor and a predecessor in the sub-list.
               Enumerator.Header.Previous_Enumerator.Next_Enumerator := Enumerator.Header.Next_Enumerator;
               Enumerator.Header.Next_Enumerator.Previous_Enumerator := Enumerator.Header.Previous_Enumerator;
            end if;
         end if;
         
         --  Now we can deallocate the enumerator's header.
         Free (Enumerator.Header);
      end if;
      
   end Finalize;

end Doubly_Linked_List;

---------------------------------------------------------------------------------------------------
--                           Licence Notice
--
--    This program is free software; you can redistribute it and/or modify
--    it under the terms of the GNU General Public License as published by
--    the Free Software Foundation; either version 2 of the License, or
--    (at your option) any later version.
--
--    This program is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--    GNU General Public License for more details.
--
--    You should have received a copy of the GNU General Public License
--    along with this program; if not, write to the Free Software
--    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--    As a special exception,  if other files  instantiate  generics from this
--    unit, or you link  this unit with other files  to produce an executable,
--    this  unit  does not  by itself cause  the resulting  executable  to  be
--    covered  by the  GNU  General  Public  License.  This exception does not
--    however invalidate  any other reasons why  the executable file  might be
--    covered by the  GNU Public License.
--
---------------------------------------------------------------------------------------------------
