---------------------------------------------------------------------------------------------------
--   Copyright (c) 2001 R.A.Matthews
--
--   See end of file for licence notice.
--
--   Revision History:
--      15 November 2001 R.A.Matthews
--         Original version.
--
---------------------------------------------------------------------------------------------------

with Ada.Streams.Stream_IO;
with Ada.Text_IO;
with Doubly_Linked_List;
pragma Elaborate_All (Doubly_Linked_List);

package body List_Examples is

   --  This package gives examples of how to use the Doubly_Linked_List package.
   --
   --  Initially we will manipulate lists of Integers:
   
   package Integer_Lists is new Doubly_Linked_List (Integer);
   
   --  Then we will show how to handle lists of different types derived from:
   
   type Variant_Type is abstract tagged null record;
   
   package Variant_Lists is new Doubly_Linked_List (Variant_Type'Class);
   
   --  For the variant type we will use:
   
   type Integer_Type is new Variant_Type with record
      I : Integer;
   end record;
   
   type String_Type is new Variant_Type with record
      S : String (1 .. 5);
   end record;
   

   --  We will print-out our results using the following shorthand:
   
   procedure PL (S : in String) renames Ada.Text_IO.Put_Line;
      
   --------------------------------------------------------------------------------------------------
   --------------------------------------------------------------------------------------------------
      
   procedure Run_Integer is
      
      --  We will load an Integer list and use an Enumerator to print it out.
      
      List_1 : Integer_Lists.List_Type;
      Enum_1 : Integer_Lists.Enumerator_Type;
      
      use Integer_Lists;
      
   begin
      PL ("");
      PL ("Here we demonstrate some features of list handling using the Integer data type");
      PL ("");
      
      PL ("At this stage our list should be empty:");
      
      if Is_Empty (List_1) then
         PL ("  The list is empty");
      else
         PL ("  The list is not empty  ** ERROR **");
      end if;
      
      PL ("And our enumerator should have no list:");
      
      if List_Exists (Enum_1) then
         PL ("  The enumerator is attached to a list  ** ERROR **");
      else
         PL ("  The enumerator is not attached to a list");
      end if;
      
      PL ("If we attach the enumerator to our list...");
      
      Enum_1 := New_Enumerator (List_1);
      
      PL ("The enumerator should now be attached:");
      
      if List_Exists (Enum_1) then
         PL ("  The enumerator is attached to a list");
      else
         PL ("  The enumerator is not attached to a list  ** ERROR **");
      end if;
      
      PL ("Using the enumerator we should find the list is empty:");
      
      if Is_Empty (Enum_1) then
         PL ("  The list is empty");
      else
         PL ("  The list is not empty  ** ERROR **");
      end if;
      
      PL ("We can load the list with some numbers");
      
      Prepend (List_1, 5);
      Prepend (List_1, 4);
      Append  (List_1, 6);
      Append  (List_1, 7);
      
      PL ("At this point the enumerator does not have a current...");
      
      if Current_Exists (Enum_1) then
         PL ("  The enumerator does have a current  ** ERROR **");
      else
         PL ("  The enumerator does not have a current");
      end if;
      
      PL ("Using the enumerator we can print-out the list elements - they should be 4  .. 7:");
      
      Goto_First (Enum_1);
      loop
         PL ("  " & Integer'Image (Current (Enum_1)));
         
         exit when Is_Last (Enum_1);   --  ****
         Goto_Next (Enum_1);
      end loop;
      
      PL ("On the other hand we can print the list elements in reverse order");
      
      Goto_Last (Enum_1);
      loop
         PL ("  " & Integer'Image (Current (Enum_1)));
         
         exit when Is_First (Enum_1);   --  *****
         Goto_Previous (Enum_1);
      end loop;
      
      declare
         Enum_2 : Enumerator_Type;
      begin
         PL ("With a second enumerator we can independently denote elements in the list");
         
         PL ("If we put the second enumerator at the start of the list:");
         
         Enum_2 := New_Enumerator (List_1);
         
         PL ("Both enumerators should be on the first element:");
         
         PL ("  The first  enumerator: " & Integer'Image (Current (Enum_1)));
         PL ("  The second enumerator: " & Integer'Image (Current (Enum_2)));
         
         PL ("So the enumerators should be equal:");
         
         if Enum_1 = Enum_2 then
            PL ("  The enumerators are equal");
         else
            PL ("  The enumerators are not equal  ** ERROR **");
         end if;
         
         PL ("If we put the second enumerator at the end of the list:");
         
         Goto_Last (Enum_2);
         
         PL ("The enumerators should be at opposite ends of the list:");
         
         PL ("  The first  enumerator: " & Integer'Image (Current (Enum_1)));
         PL ("  The second enumerator: " & Integer'Image (Current (Enum_2)));
         
         PL ("So the enumerators should be unequal:");
         
         if Enum_1 = Enum_2 then
            PL ("  The enumerators are equal  ** ERROR **");
         else
            PL ("  The enumerators are not equal");
         end if;
         
         PL ("If we modify the list via one enumerator, the other enumerator should see the change:");
         
         Goto_Previous (Enum_2);
         Goto_Previous (Enum_2);
         Goto_Next     (Enum_1);
         
         PL ("Both enumerators should be denoting 5:");
         
         PL ("  The first  enumerator: " & Integer'Image (Current (Enum_1)));
         PL ("  The second enumerator: " & Integer'Image (Current (Enum_2)));
         
         PL ("Now delete the list element via the first enumerator:");
         
         Delete (Enum_1);
         
         PL ("Both enumerators should now not have a current element:");
         
         if Current_Exists (Enum_1) then
            PL ("  Via the first enumerator, the element still exists  ** ERROR **");
         else
            PL ("  Via the first enumerator, the element does not exist");
         end if;
         
         if Current_Exists (Enum_2) then
            PL ("  Via the second enumerator, the element still exists  ** ERROR **");
         else
            PL ("  Via the second enumerator, the element does not exist");
         end if;
         
         PL ("Using a third enumerator the list should now contain 4, 6 and 7:");
         
         declare
            Enum_3 : Enumerator_Type;
         begin
            Enum_3 := New_Enumerator (List_1);
            
            loop
               PL ("  " & Integer'Image (Current (Enum_3)));
               
               exit when Is_Last (Enum_3);   --  ****
               Goto_Next (Enum_3);
            end loop;
         end;
         
         PL ("If we move both 1st and 2nd enumerators forwards, they should both denote 6:");
         
         Goto_Next (Enum_1);
         Goto_Next (Enum_2);
         
         PL ("  The first  enumerator: " & Integer'Image (Current (Enum_1)));
         PL ("  The second enumerator: " & Integer'Image (Current (Enum_2)));
         
         PL ("We can use the enumerators to insert into the list, giving 4, 9, 6, 8, 7:");
         
         Insert (Enum_1, 9, Before_Current);
         Insert (Enum_2, 8, After_Current);
         
         declare
            Enum_3 : Enumerator_Type;
         begin
            Enum_3 := New_Enumerator (List_1);
            
            loop
               PL ("  " & Integer'Image (Current (Enum_3)));
               
               exit when Is_Last (Enum_3);   --  ****
               Goto_Next (Enum_3);
            end loop;
         end;
         
         PL ("Via assignment we can quickly set a third enumerator the same as the 1st enumerator (to denote 9):");
         
         declare
            Enum_3 : Enumerator_Type;
         begin
            Enum_3 := Enum_1;
            
            PL ("  The first enumerator: " & Integer'Image (Current (Enum_1)));
            PL ("  The third enumerator: " & Integer'Image (Current (Enum_3)));
         end;
         
         PL ("A second list can be quickly set via assignment:");
         
         declare
            List_2 : List_Type;
         begin
            List_2 := List_1;
            
            PL ("The second list should now contain 4, 9, 6, 8, 7:");
            
            declare
               Enum_3 : Enumerator_Type;
            begin
               Enum_3 := New_Enumerator (List_2);
            
               loop
                  PL ("  " & Integer'Image (Current (Enum_3)));
                  
                  exit when Is_Last (Enum_3);   --  ****
                  Goto_Next (Enum_3);
               end loop;
            end;
            
            PL ("And the lists should be equal:");
            
            if List_1 = List_2 then
               PL ("  The lists are equal");
            else
               PL ("  The lists are not equal  ** ERROR **");
            end if;
            
            PL ("We can delete all entries in the second list without affecting the first list:");
            
            Delete_All (List_2);
            
            if Is_Empty (List_2) then
               PL ("  The 2nd list is empty");
            else
               PL ("  The 2nd list is not empty  ** ERROR **");
            end if;
            
            PL ("The 1st list should still contain 4, 9, 6, 8, 7:");
            
            declare
               Enum_3 : Enumerator_Type;
            begin
               Enum_3 := New_Enumerator (List_1);
            
               loop
                  PL ("  " & Integer'Image (Current (Enum_3)));
                  
                  exit when Is_Last (Enum_3);   --  ****
                  Goto_Next (Enum_3);
               end loop;
            end;
            
            PL ("And the lists should be unequal:");
            
            if List_1 = List_2 then
               PL ("  The lists are equal  ** ERROR **");
            else
               PL ("  The lists are not equal");
            end if;
            
            PL ("If we attach the 2nd enumerator to the 2nd list, then let the 2nd list go out of scope,");
            PL ("the 2nd enumerator should be aware of the change:");
            
            Enum_2 := New_Enumerator (List_2);
            
            if List_Exists (Enum_2) then
               PL ("  With the 2nd list still in scope, the 2nd enumerator is attached");
            else
               PL ("  With the 2nd list still in scope, the 2nd enumerator is not attached  ** ERROR **");
            end if;
         end;
         if List_Exists (Enum_2) then
            PL ("  With the 2nd list no longer in scope, the 2nd enumerator is attached  ** ERROR **");
         else
            PL ("  With the 2nd list no longer in scope, the 2nd enumerator is not attached");
         end if;
         
         PL ("We can modify the 1st list to contain 5, 10, 7, 9, 8:");
         
         declare
            Enum_3 : Enumerator_Type;
         begin
            Enum_3 := New_Enumerator (List_1);
            loop
               Update (Enum_3, (Current (Enum_3) + 1));
               
               exit when Is_Last (Enum_3);   --  ****
               Goto_Next (Enum_3);
            end loop;
            
            Goto_First (Enum_3);
            loop
               PL ("  " & Integer'Image (Current (Enum_3)));
                  
               exit when Is_Last (Enum_3);   --  ****
               Goto_Next (Enum_3);
            end loop;
         end;
      end;
   end Run_Integer;
      
   --------------------------------------------------------------------------------------------------
   --------------------------------------------------------------------------------------------------
      
   procedure Run_Variant is
      
      List_1 : Variant_Lists.List_Type;
      Enum_1 : Variant_Lists.Enumerator_Type;
            
      use Variant_Lists;
   begin
      PL ("");
      PL ("Here we demonstrate some features of list handling using the variant data type");
      PL ("");
      
      Append (List_1, Integer_Type'(I => 123));
      Append (List_1, String_Type'(S => "abcde"));
            
      PL ("The list should contain the integer 123 followed by the string 'abcde':");
            
      Enum_1 := New_Enumerator (List_1);
            
      loop
         declare
            Value : Variant_Type'Class := Current (Enum_1);
         begin
            if Value in Integer_Type then
               PL (" Integer value: " & Integer'Image (Integer_Type (Value).I));
                  
            elsif Value in String_Type then
               PL ("  String value: " & String_Type (Value).S);
            else
               PL ("** Invalid list element type **");
            end if;
         end;
         exit when Is_Last (Enum_1);   --  ***********
         Goto_Next (Enum_1);
      end loop;
            
      PL ("Using update we can change the list to the string 'fghij' following by integer 456:");
            
      Update (Enum_1, Integer_Type'(I => 456));
      Goto_Previous (Enum_1);
      Update (Enum_1, String_Type'(S => "fghij"));
            
      Goto_First (Enum_1);
      loop
         declare
            Value : Variant_Type'Class := Current (Enum_1);
         begin
            if Value in Integer_Type then
               PL (" Integer value: " & Integer'Image (Integer_Type (Value).I));
                  
            elsif Value in String_Type then
               PL ("  String value: " & String_Type (Value).S);
            else
               PL ("** Invalid list element type **");
            end if;
         end;
         exit when Is_Last (Enum_1);
         Goto_Next (Enum_1);
      end loop;
         
      PL ("Using stream facilities we can write the list to a temporary file");
            
      declare
         Stream_File : Ada.Streams.Stream_IO.File_Type;
         Stream      : Ada.Streams.Stream_IO.Stream_Access;
      begin
         Ada.Streams.Stream_IO.Create (File => Stream_File);
         Stream := Ada.Streams.Stream_IO.Stream (Stream_File);
              
         List_Type'Write (Stream, List_1);
               
         declare
            List_2 : List_Type;
         begin
            Ada.Streams.Stream_IO.Reset (File => Stream_File, Mode => Ada.Streams.Stream_IO.In_File);
               
            List_Type'Read (Stream, List_2);
                  
            PL ("We have just read in the list from file into a 2nd list, which should now contain");
            PL ("the string 'fghij' following by integer 456:");
                  
            declare
               Enum_2 : Enumerator_Type;
            begin
               Enum_2 := New_Enumerator (List_2);
                     
               loop
                  declare
                     Value : Variant_Type'Class := Current (Enum_2);
                  begin
                     if Value in Integer_Type then
                        PL (" Integer value: " & Integer'Image (Integer_Type (Value).I));
                  
                     elsif Value in String_Type then
                        PL ("  String value: " & String_Type (Value).S);
                     else
                        PL ("** Invalid list element type **");
                     end if;
                  end;
                  exit when Is_Last (Enum_2);
                  Goto_Next (Enum_2);
               end loop;
            end;
            
            PL ("We can write out an empty list and read it back in:");
            
            Delete_All (List_1);
            
            Ada.Streams.Stream_IO.Reset (File => Stream_File, Mode => Ada.Streams.Stream_IO.Out_File);
            
            List_Type'Write (Stream, List_1);
            
            Ada.Streams.Stream_IO.Reset (File => Stream_File, Mode => Ada.Streams.Stream_IO.In_File);
               
            List_Type'Read (Stream, List_2);
            
            PL ("We have just read the list into the 2nd list - it should now be empty:");
            
            if Is_Empty (List_2) then
               PL ("  The 2nd list is empty");
            else
               PL ("  The 2nd list is not empty  ** ERROR **");
            end if;
         end;
         Ada.Streams.Stream_IO.Close (File => Stream_File);
      end;

   end Run_Variant;

   --------------------------------------------------------------------------------------------------
   --------------------------------------------------------------------------------------------------
   
   
end List_Examples;

---------------------------------------------------------------------------------------------------
--                           Licence Notice
--
--    This program is free software; you can redistribute it and/or modify
--    it under the terms of the GNU General Public License as published by
--    the Free Software Foundation; either version 2 of the License, or
--    (at your option) any later version.
--
--    This program is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--    GNU General Public License for more details.
--
--    You should have received a copy of the GNU General Public License
--    along with this program; if not, write to the Free Software
--    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
---------------------------------------------------------------------------------------------------
